// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/neck_turn_to_ball_or_scan.h>

#include "bhv_danger_area_tackle.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_DangerAreaTackle::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_DangerAreaTackle"
                        ,__FILE__, __LINE__ );

    if ( ! agent->world().existKickableOpponent()
         || agent->world().self().tackleProbability() < M_min_probability )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: failed"
                            ,__FILE__, __LINE__ );
        return false;
    }

    if ( agent->world().self().pos().absY() > rcsc::ServerParam::i().goalHalfWidth() + 5.0 )
    {
        // out of goal
        double power = rcsc::ServerParam::i().maxPower();
        if ( agent->world().self().body().abs() < 10.0 )
        {
            // nothing to do
        }
        else if ( agent->world().self().body().abs() > 170.0 )
        {
            power *= -1.0;
        }
        else if ( agent->world().self().body().degree() * agent->world().self().pos().y < 0.0 )
        {
            power *= -1.0;
        }
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: out of goal width"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "tackle(1)" );
        agent->doTackle( power );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return true;
    }
    else
    {
        // within goal width
        double power_sign = 0.0;
        if ( agent->world().self().pos().y > 0.0 )
        {
            if ( 0.0 < agent->world().self().body().degree()
                 && agent->world().self().body().degree() < 90.0 )
            {
                power_sign = 1.0;
            }
            else if ( -180.0 < agent->world().self().body().degree()
                      && agent->world().self().body().degree() < -90.0 )
            {
                power_sign = -1.0;
            }
        }
        else
        {
            if ( -90.0 < agent->world().self().body().degree()
                 && agent->world().self().body().degree() < 0.0 )
            {
                power_sign = 1.0;
            }
            else if ( 90.0 < agent->world().self().body().degree()
                      && agent->world().self().body().degree() < 180.0 )
            {
                power_sign = -1.0;
            }

        }
        if ( power_sign != 0.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: power_sigh = %f"
                                ,__FILE__, __LINE__,
                                power_sign );
            agent->debugClient().addMessage( "tackle(%f)", power_sign );
            agent->doTackle( power_sign * rcsc::ServerParam::i().maxPower() );
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            return true;
        }
    }

    return false;
}
