// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_basic_move.h"
#include "bhv_defender_get_ball.h"
#include "bhv_danger_area_tackle.h"

#include "bhv_defender_basic_block_move.h"

/*-------------------------------------------------------------------*/
/*!
  execute action
*/
bool
Bhv_DefenderBasicBlockMove::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_DefenderBasicBlock"
                        ,__FILE__, __LINE__ );
    //////////////////////////////////////////////////////////////////
    // tackle
    if ( Bhv_DangerAreaTackle().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: tackle"
                            ,__FILE__, __LINE__ );
        return true;
    }

    //////////////////////////////////////////////////////////////////
    // get ball
    if ( Bhv_DefenderGetBall( M_home_pos ).execute( agent ) )
    {
        return true;
    }

    //////////////////////////////////////////////////////////////////
    // positioning

    const double ball_xdiff
        = agent->world().ball().pos().x - agent->world().self().pos().x;

    if ( ball_xdiff > 10.0
         && agent->world().getInterceptTable()->isOurTeamBallPossessor() )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: ball is front and our team keep ball"
                            ,__FILE__, __LINE__ );
        Bhv_BasicMove( M_home_pos ).execute( agent );
        return true;
    }

    double dash_power = get_dash_power( agent, M_home_pos );

    agent->debugClient().addMessage( "DefMove%.0f", dash_power );
    agent->debugClient().setTarget( M_home_pos );
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: go home. power= %f"
                        ,__FILE__, __LINE__,
                        dash_power );
    double dist_thr = agent->world().ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.5 ) dist_thr = 1.5;
    if ( ! rcsc::Body_GoToPoint( M_home_pos,
                                 dist_thr,
                                 dash_power
                                 ).execute( agent ) )
    {
        rcsc::Vector2D body_point( -100.0,
                                   agent->world().self().pos().y );
        if ( agent->world().ball().angleFromSelf().abs() < 80.0 )
        {
            body_point.x = 100.0;
        }
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }

    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );

    return true;
}

/*-------------------------------------------------------------------*/
/*!
  static method
*/
double
Bhv_DefenderBasicBlockMove::get_dash_power( const rcsc::PlayerAgent * agent,
                                            const rcsc::Vector2D & home_pos )
{
    static bool S_recover_mode = false;

    const rcsc::WorldModel & wm = agent->world();

    if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.5 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.85 )
    {
        S_recover_mode = false;
    }


    const double ball_xdiff
        = wm.ball().pos().x - wm.self().pos().x;

    const rcsc::PlayerType & mytype = wm.self().playerType();
    const double my_inc = mytype.staminaIncMax() * wm.self().recovery();

    double dash_power;
    if ( S_recover_mode )
    {
        if ( wm.getDefenseLineX() > wm.self().pos().x )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: correct DF line & recover"
                                ,__FILE__, __LINE__ );
            dash_power = my_inc;
        }
        else if ( ball_xdiff > 20.0 )
        {
            dash_power = std::max( 0.0, my_inc - 20.0 );
        }
        else if ( ball_xdiff > 10.0 )
        {
            dash_power = std::max( 0.0, my_inc - 10.0 );
        }
        else if ( ball_xdiff > 5.0 )
        {
            dash_power = rcsc::ServerParam::i().maxPower();
            dash_power *= 0.7;
        }
        else
        {
            dash_power = rcsc::ServerParam::i().maxPower();
        }

        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: recover mode dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }
    else
    {
        if ( wm.getDefenseLineX() > wm.self().pos().x )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: correct DF line"
                                ,__FILE__, __LINE__ );
            dash_power = rcsc::ServerParam::i().maxPower();
        }
        else if ( ball_xdiff > 20.0 )
        {
            dash_power = my_inc;
        }
        else if ( ball_xdiff > 10.0 )
        {
            dash_power = rcsc::ServerParam::i().maxPower();
            dash_power *= 0.6;
        }
        else if ( ball_xdiff > 5.0 )
        {
            dash_power = rcsc::ServerParam::i().maxPower();
            dash_power *= 0.85;
        }
        else
        {
            dash_power = rcsc::ServerParam::i().maxPower();
        }
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: nomal mode dash_power= %f"
                            ,__FILE__, __LINE__,
                            dash_power );
    }

    return dash_power;
}
