// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_shoot.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include "bhv_cross_to_center.h"
#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_keep_shoot_chance.h"

#include "bhv_basic_tackle.h"

#include "bhv_basic_move.h"
#include "bhv_defender_basic_block_move.h"
#include "bhv_defender_get_ball.h"
#include "bhv_mark_close_opponent.h"
#include "bhv_danger_area_tackle.h"

#include "strategy.h"
#include "formation.h"
#include "formation_uva.h"

#include "role_center_back.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::execute( rcsc::PlayerAgent * agent )
{

    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().getTeammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::doKick( rcsc::PlayerAgent * agent )
{
    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 1000.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );

    switch ( Strategy::get_ball_area( agent->world().ball().pos()) ) {
    case Strategy::BA_CrossBlock:
        if ( closest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: cross block area. danger"
                                ,__FILE__, __LINE__ );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                rcsc::Body_AdvanceBall().execute( agent );
            }
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
        if ( closest_opp_dist < 3.0 )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s:%d: stopper or danger area."
                                ,__FILE__, __LINE__ );
            if ( ! rcsc::Body_Pass().execute( agent ) )
            {
                rcsc::Body_ClearBall().execute( agent );
            }
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicDefensiveKick().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_Cross:
        Bhv_CrossToCenter().execute( agent );
        break;
    case Strategy::BA_ShootChance:
        if ( rcsc::Body_Shoot().execute( agent ) )
        {
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_KeepShootChance().execute( agent );
        }
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        rcsc::Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::Vector2D base_pos = agent->world().ball().pos();
    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, agent->world().getOffsideLineX() - 1.0 );
    }

    ///// delete here, if you do not use SBSP formation ////////
    if ( uva::Formations::instance().valid() )
    {
    if ( agent->world().getOurSide() == 1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() > 0 )
     {
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else    
      {

        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
      else
      {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else    
      {

        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      }
     }
    if( agent->world().getOurSide() == -1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() < 0 )
     {
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else    
      {

        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
      else
      {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if (agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else    
      {

        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      }
     }
    }
    ///////////////////////////////////////////////////////

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: HOME POSITION=(%.3f, %.3f)"
                        ,__FILE__, __LINE__,
                        home_pos.x, home_pos.y );

    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_CrossBlock:
        doMarkMove( agent, home_pos );
        break;
    case Strategy::BA_Stopper:
#if 1
        if ( agent->world().ball().pos().x > home_pos.x + 1.0 )
        {
            agent->debugClient().addMessage( "StopperMove(1)" );
            doBasicMove( agent, home_pos );
        }
        else
#endif
            if ( agent->world().ball().pos().absY()
                 < rcsc::ServerParam::i().goalHalfWidth() + 2.0 )
            {
                agent->debugClient().addMessage( "StopperMove(2)" );
                doStopperMove( agent, home_pos );
            }
            else
            {
                agent->debugClient().addMessage( "StopperMove(3)" );
                doMarkMove( agent, home_pos );
            }
        break;
    case Strategy::BA_Danger:
        if ( agent->world().ball().pos().absY()
             < rcsc::ServerParam::i().goalHalfWidth() )
        {
            agent->debugClient().addMessage( "DangerMove(1)" );
            doStopperMove( agent, home_pos );
        }
        else
        {
            agent->debugClient().addMessage( "DangerMove(2)" );
            doMarkMove( agent, home_pos );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
        doBasicMove( agent, home_pos );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball pos"
                            ,__FILE__, __LINE__ );
        Bhv_BasicMove( home_pos ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::doBasicMove( rcsc::PlayerAgent * agent,
                             const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doBasicMove"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();

    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.9, 50.0 ).execute( agent ) )
    {
        return;
    }

    //--------------------------------------------------
    // check intercept chance
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: doBasicMove. intercept cycle. self=%d, mate=%d, opp=%d"
                        ,__FILE__, __LINE__,
                        self_min, mate_min, opp_min );

    if ( ( ! wm.existKickableTeammate()
           && self_min <= 3
           && opp_min >= 4
           )
         || ( ! wm.existKickableTeammate()
              && self_min <= 2 )
         || ( self_min < 20
              && ( self_min < mate_min
                   || ( self_min < mate_min + 3
                        && mate_min >= 4 ) )
              && ( self_min < opp_min
                   && opp_min > 1 ) )
         )
    {
        // chase ball
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doBasicMove intercept"
                            ,__FILE__ );
        agent->debugClient().addMessage( "BasicMove:Intercept" );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    /*--------------------------------------------------------*/
    // go to home position

    // ball is in very safety area (very front & our ball)
    if ( wm.ball().pos().x > wm.self().pos().x + 10.0
         && wm.getInterceptTable()->isOurTeamBallPossessor() )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball is front and our team own ball"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage("CB:safety");
        Bhv_BasicMove( home_pos ).execute( agent );
        return;
    }

    // decide wait position
    // adjust to the defence line
    rcsc::Vector2D target_point = home_pos;
    if ( wm.self().pos().x > home_pos.x
         && wm.ball().pos().x > wm.self().pos().x )
    {
        // make static line
        // static line
        // 0, -10, -20, ...
        double tmpx = std::floor( wm.ball().pos().x * 0.1 ) * 10.0 - 3.0;
        if ( tmpx > 0.0 ) tmpx = 0.0;
        if ( tmpx > home_pos.x + 5.0 ) tmpx = home_pos.x + 5.0;
        target_point.x = tmpx;
        if ( wm.getDefenseLineX() < target_point.x - 1.0 )
        {
            target_point.x = wm.getDefenseLineX() - 2.0;
        }

        target_point.y = home_pos.y;
        double home_y_diff = wm.ball().pos().y - home_pos.y;
        if ( wm.ball().pos().absY() < home_pos.absY() )
        {
            if ( std::fabs( home_y_diff ) < 3.0 )
            {
                target_point.y = wm.ball().pos().y;
            }
            else
            {
                target_point.y += home_y_diff * 0.4;
            }
        }
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: correct target point to wait ball. y-correct=%.3f"
                            ,__FILE__, __LINE__,
                            home_y_diff );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: check block center. bpos(%.1f %.1f ) bvelx = %.2f"
                        ,__FILE__, __LINE__,
                        wm.ball().pos().x, wm.ball().pos().y,
                        wm.ball().vel().x );
    if ( wm.ball().pos().absY() < 5.0
         && wm.ball().pos().x < wm.self().pos().x + 1.0
         && wm.ball().pos().x > -42.0
         && ( wm.ball().vel().x < -0.7
              || wm.existKickableOpponent()
              || opp_min <= 1 )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: correct target point to block center"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "BlockCenter" );
        target_point.x = -45.0;
        target_point.y = 0.0;
    }

    // decide dash power
    double dash_power
        = Bhv_DefenderBasicBlockMove::get_dash_power( agent, target_point );

    // decide threshold
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: go to home. power=%.1f"
                        ,__FILE__, __LINE__,
                        dash_power );
    agent->debugClient().addMessage( "CB:basic %.0f", dash_power );
    agent->debugClient().setTarget( target_point );

    if ( wm.ball().pos().x < -35.0 )
    {
        if ( wm.self().stamina() > rcsc::ServerParam::i().staminaMax() * 0.6
             && wm.self().pos().x < target_point.x - 4.0
             && wm.self().pos().dist( target_point ) > dist_thr )
        {
            rcsc::Bhv_GoToPointLookBall( target_point,
                                         dist_thr,
                                         dash_power
                                         ).execute( agent );
            return;
        }
    }

    if ( ! rcsc::Body_GoToPoint( target_point,
                                 dist_thr,
                                 dash_power,
                                 100, // cycle
                                 false, // no back
                                 true, // stamina save
                                 25.0 // dir_thr
                                 ).execute( agent ) )
    {
        rcsc::Vector2D body_point( wm.self().pos().x, 0.0 );
        body_point.y = ( wm.ball().pos().y < wm.self().pos().y
                         ? -40.0
                         : 40.0 );
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }

    if ( wm.existKickableOpponent()
         || opp_min <= 3 )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleCenterBack::doStopperMove( rcsc::PlayerAgent * agent,
                               const rcsc::Vector2D &home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: doStopperMove"
                        ,__FILE__, __LINE__ );

    ////////////////////////////////////////////////////////
    // search mark target opponent
    if ( doDangerGetBall( agent, home_pos ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doStopperMove. done doDangetGetBall"
                            ,__FILE__, __LINE__ );
        return;
    }
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: no gettable point"
                        ,__FILE__, __LINE__ );
    Bhv_DefenderBasicBlockMove( home_pos ).execute( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/

void
RoleCenterBack::doMarkMove( rcsc::PlayerAgent * agent,
                            const rcsc::Vector2D& home_pos )
{
    //-----------------------------------------------
    // tackle
    if ( Bhv_DangerAreaTackle().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. done tackle"
                            ,__FILE__ );
        return;
    }

    ////////////////////////////////////////////////////////

    if ( doDangerGetBall( agent, home_pos ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. done get ball"
                            ,__FILE__  );
        return;
    }

    const rcsc::WorldModel & wm = agent->world();

    ////////////////////////////////////////////////////////
    // search mark target opponent

    const rcsc::PlayerObject * mark_target
        = wm.getOpponentNearestTo( home_pos, 1, NULL );

    ////////////////////////////////////////////////////////
    // check candidate opponent
    if ( ! mark_target
         || mark_target->posCount() > 0
         || mark_target->distFromBall() < rcsc::ServerParam::i().defaultKickableArea() + 0.5
         || mark_target->distFromSelf() > 7.0
         || mark_target->pos().dist( home_pos ) > 6.0
         || mark_target->pos().x > 39.0
         )
    {
        // not found
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. mark not found. change to basic move"
                            ,__FILE__ );
        agent->debugClient().addMessage( "MarkNotFound" );
        doBasicMove( agent, home_pos );
        return;
    }

    ////////////////////////////////////////////////////////
    // check teammate closer than self
    {
        double marker_dist = 100.0;
        const rcsc::PlayerObject * marker
            = wm.getTeammateNearestTo( mark_target->pos(),
                                       30,
                                       &marker_dist );
        if ( marker && marker_dist < mark_target->distFromSelf() )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doMarkMove. exist other marker"
                                ,__FILE__ );
            Bhv_DefenderBasicBlockMove( home_pos ).execute( agent );
            return;
        }
    }

    ////////////////////////////////////////////////////////
    // set target point
    rcsc::Vector2D block_point = mark_target->pos();
    block_point += mark_target->vel() / 0.6;
    block_point.x -= 0.9;
    block_point.y += ( mark_target->pos().y > wm.ball().pos().y
                       ? -0.6
                       : 0.6 );
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: doMarkMove. mark point(%.1f %.1f)"
                        ,__FILE__,
                        block_point.x, block_point.y );

    if ( block_point.x > wm.ball().pos().x + 5.0 )
    {
        // not found
        agent->debugClient().addMessage( "MarkToBasic" );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. (mark_point.x - ball.x) X diff is big"
                            ,__FILE__ );
        doBasicMove( agent, home_pos );
        return;
    }

    double dash_power = rcsc::ServerParam::i().maxPower();
    double x_diff = block_point.x - wm.self().pos().x;

    if ( x_diff > 20.0 )
    {
        //dash_power *= 0.5;
        dash_power = wm.self().playerType().staminaIncMax() * wm.self().recovery();
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. power change(1) X diff=%.2f dash_power=%.1f "
                            ,__FILE__,
                            x_diff, dash_power );
    }
    else if ( x_diff > 10.0 )
    {
        dash_power *= 0.7;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. power change(2) X diff=%.2f dash_power=%.1f "
                            ,__FILE__,
                            x_diff, dash_power );
    }
    else if ( wm.ball().pos().dist( block_point ) > 20.0 )
    {
        dash_power *= 0.6;
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. mark point is far. dash power=%.1f "
                            ,__FILE__,
                            dash_power );
    }
    agent->debugClient().addMessage( "MarkDashPower%.0f", dash_power );
    agent->debugClient().setTarget( block_point );

    double dist_thr = wm.ball().distFromSelf() * 0.05;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    if ( wm.self().pos().x < block_point.x
         && wm.self().pos().dist( block_point ) < dist_thr )
    {
        rcsc::Vector2D face_point( -52.5, wm.self().pos().y );
        if ( wm.ball().pos().x < wm.self().pos().x - 5.0 )
        {
            face_point.x = 0.0;
        }
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. already there. face to(%.1f %.1f)"
                            ,__FILE__,
                            face_point.x, face_point.y );
        rcsc::Body_TurnToPoint( face_point ).execute( agent );
        if ( wm.existKickableOpponent()
             || wm.ball().distFromSelf() > 15.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return;
    }

    if ( wm.self().pos().dist( block_point ) > 3.0 )
    {
        agent->debugClient().addMessage( "MarkForwardDash" );
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. forward move"
                            ,__FILE__ );
        rcsc::Body_GoToPoint( block_point, dist_thr, dash_power
                              ).execute( agent );
        if ( wm.existKickableOpponent()
             || wm.ball().distFromSelf() > 15.0 )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
        }
        else
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        }
        return;
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: doMarkMove. ball looking move"
                        ,__FILE__ );

    rcsc::Bhv_GoToPointLookBall( block_point,
                                 dist_thr,
                                 dash_power
                                 ).execute( agent );

    const rcsc::PlayerObject * ball_owner = wm.getOpponentNearestToBall( 5 );

    if ( ! ball_owner
         || ball_owner->distFromBall() > 1.2 + wm.ball().distFromSelf() * 0.05 )
    {
        if ( mark_target && mark_target->unum() < 0 )
        {
            rcsc::AngleDeg angle_diff
                = agent->effector().queuedNextAngleFromBody( mark_target->pos()
                                                             + mark_target->vel() );
            if ( angle_diff.abs() < 100.0 )
            {
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s: doMarkMove. check unknown mark target"
                                    ,__FILE__ );
                agent->setNeckAction( new rcsc::Neck_TurnToPoint( mark_target->pos()
                                                                  + mark_target->vel() ) );
                return;
            }
        }
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doMarkMove. look ball or field scan"
                            ,__FILE__ );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
RoleCenterBack::doDangerGetBall( rcsc::PlayerAgent * agent,
                                 const rcsc::Vector2D& home_pos )
{
    //////////////////////////////////////////////////////////////////
    // tackle
    if ( Bhv_DangerAreaTackle().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doDangetGetBall done tackle"
                            ,__FILE__ );
        return true;
    }

    const rcsc::WorldModel & wm = agent->world();

    //--------------------------------------------------
    // intercept
    if ( ! wm.existKickableTeammate()
         && wm.ball().pos().x < -36.0
         && wm.ball().pos().dist( home_pos ) < 7.0
         && wm.ball().pos().absY() < 9.0
         )
    {
        if ( wm.existKickableOpponent() )
        {
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doDangetGetBall attack to ball owner"
                                ,__FILE__ );
            if ( wm.self().pos().x > wm.ball().pos().x + 0.5 )
            {
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s: doDangetGetBall  ball is behind"
                                    ,__FILE__ );
                return false;
            }

            rcsc::Vector2D opp_pos = wm.getOpponentsFromBall().front()->pos();
            rcsc::Vector2D opp_vel = wm.getOpponentsFromBall().front()->vel();
            opp_pos.x -= 0.2;
            if ( opp_vel.x < -0.1 )
            {
                opp_pos.x -= 0.5;
            }
            if ( opp_pos.x > wm.self().pos().x
                 && fabs(opp_pos.y - wm.self().pos().y) > 0.8 )
            {
                opp_pos.x = wm.self().pos().x;
            }

            agent->debugClient().addMessage( "DangerAttackOpp" );
            agent->debugClient().setTarget( opp_pos );
            rcsc::dlog.addText( rcsc::Logger::ROLE,
                                "%s: doDangetGetBall (%.2f %.2f)"
                                ,__FILE__,
                                opp_pos.x, opp_pos.y );
            rcsc::Body_GoToPoint( opp_pos,
                                  0.1,
                                  rcsc::ServerParam::i().maxPower()
                                  ).execute( agent );
            if ( wm.existKickableOpponent() )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            }
            return true;
        }
        else
        {
            int self_min = wm.getInterceptTable()->getSelfReachCycle();

            if ( self_min < 10 )
            {
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s: doDangetGetBall. intercept. reach cycle = %d"
                                    ,__FILE__,
                                    self_min );
                agent->debugClient().addMessage( "DangerGetBall" );

                rcsc::Body_Intercept().execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
                return true;
            }
        }
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s: doDangetGetBall. never get"
                            ,__FILE__ );
    }

    return false;
}
