// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <rcsc/soccer_math.h>
#include <rcsc/param/server_param.h>

#include <rcsc/player/logger.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/player_config.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_shoot.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_scan_field.h>

#include "bhv_cross_to_center.h"
#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_keep_shoot_chance.h"

#include "bhv_basic_tackle.h"

#include "bhv_basic_move.h"
#include "bhv_attacker_offensive_move.h"
#include "bhv_mark_close_opponent.h"
#include "bhv_press.h"

#include "strategy.h"
#include "formation.h"
#include "formation_uva.h"

#include "role_offensive_half.h"

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::execute( rcsc::PlayerAgent * agent )
{
    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().getTeammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
    const double closest_opp_dist
        = ( agent->world().getOpponentsFromSelf().empty()
            ? 1000.0
            : agent->world().getOpponentsFromSelf().front()->distFromSelf() );

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
        if ( closest_opp_dist < rcsc::ServerParam::i().tackleDist() )
        {
            rcsc::Body_AdvanceBall().execute( agent );
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicOffensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_Danger:
        if ( closest_opp_dist < rcsc::ServerParam::i().tackleDist() )
        {
            rcsc::Body_ClearBall().execute( agent );
            agent->setNeckAction( new rcsc::Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
        doDribbleAttackKick( agent );
        break;
    case Strategy::BA_OffMidField:
        agent->debugClient().addMessage( "OffMid" );
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_Cross:
        agent->debugClient().addMessage( "AreaCross" );
        Bhv_CrossToCenter().execute( agent );
        break;
    case Strategy::BA_ShootChance:
        doShootChanceKick( agent );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        rcsc::Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::Vector2D base_pos = agent->world().ball().pos();
    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, agent->world().getOffsideLineX() - 1.0 );
    }

/*    ///// delete here, if you do not use SBSP formation ////////

    if ( uva::Formations::instance().valid() )
    {
    if ( agent->world().getOurSide() == 1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() > 0 )
     {   
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    }
    if ( agent->world().getOurSide() == -1 )
    {
     if ( agent->world().getGameMode().getScoreLeft() - agent->world().getGameMode().getScoreRight() < 0 )
     {   
      if ( agent->world().ball().pos().x > 35.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < 0.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_SDEFENSIV );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    else
    {
      if ( agent->world().ball().pos().x > 20.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_334_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else if ( agent->world().ball().pos().x < -15.0 )
      {
        uva::Formations::instance().setFormation( uva::FT_523_DEFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
      else
      {
        uva::Formations::instance().setFormation( uva::FT_433_OFFENSIVE );
        home_pos = uva::Formations::instance()
            .getStrategicPosition( agent->config().playerNumber() - 1,
                                   base_pos,
                                   agent->world().getOffsideLineX() - 1.0 );
      }
     }
    }
    }
*/    ///////////////////////////////////////////////////////


    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: HOME POSITION=(%.3f, %.3f)"
                        ,__FILE__, __LINE__,
                        home_pos.x, home_pos.y );

    switch ( Strategy::get_ball_area(base_pos) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        agent->debugClient().addMessage( "DefMove" );
        doDefensiveMove( agent, home_pos );
        break;
    case Strategy::BA_DribbleAttack:
        agent->debugClient().addMessage( "DribAttMove" );
        if ( ! doShootSupportMove( agent, home_pos ) )
        {
            //Act_BasicMove( home_pos ).execute( agent );
            doOffensiveMove( agent, home_pos );
        }
        break;
    case Strategy::BA_OffMidField:
        agent->debugClient().addMessage( "OffMove" );
        doOffensiveMove( agent, home_pos );
        break;
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
        agent->debugClient().addMessage( "ShotAreaMove" );
        doShootAreaMove( agent, home_pos );
        break;
    default:
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: unknown ball area"
                            ,__FILE__, __LINE__ );
        Bhv_BasicMove( home_pos ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doDribbleAttackKick( rcsc::PlayerAgent * agent )
{
    agent->debugClient().addMessage( "DribAtt" );
    Bhv_BasicOffensiveKick().execute( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doShootChanceKick( rcsc::PlayerAgent * agent )
{
    agent->debugClient().addMessage( "ShootChance" );
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: keep shoot chance."
                        ,__FILE__, __LINE__ );
    Bhv_KeepShootChance().execute( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doOffensiveMove( rcsc::PlayerAgent * agent,
                                    const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doOffensiveMove"
                        ,__FILE__, __LINE__ );
    //----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    const rcsc::WorldModel & wm = agent->world();
    //----------------------------------------------
    // intercept
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();

    if ( self_min < mate_min )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doOffensiveMove. I am fastest in teammate"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "OffGetBall(1)" );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( wm.existKickableOpponent()
         && wm.ball().pos().dist( home_pos ) < 12.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: doOffensiveMove. I am fastest in teammate"
                            ,__FILE__, __LINE__ );
        rcsc::Vector2D attack_point = wm.getOpponentsFromBall().front()->pos();
        attack_point.x -= 0.5;
        double dash_power = rcsc::ServerParam::i().maxPower();
        if ( wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.6 )
        {
            dash_power *= wm.self().stamina() / rcsc::ServerParam::i().staminaMax();
        }
        agent->debugClient().addMessage( "OffAttackBall" );
        agent->debugClient().setTarget( attack_point );
        rcsc::Body_GoToPoint( attack_point, 0.5, dash_power ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( wm.getInterceptTable()->isOurTeamBallPossessor() )
    {
        if ( wm.ball().distFromSelf() < 10.0 )
        {
            if ( wm.getInterceptTable()->isSelfFastestPlayer() )
            {
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s:%d: doOffensiveMove. ball owner is our. I am fastest.  interept"
                                    ,__FILE__, __LINE__ );
                agent->debugClient().addMessage( "OffGetBall(2)" );
                rcsc::Body_Intercept().execute( agent );
                if ( wm.ball().distFromSelf() < rcsc::ServerParam::i().visibleDistance() )
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                }
                else
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
                }
                return;
            }
            else
            {
                rcsc::dlog.addText( rcsc::Logger::ROLE,
                                    "%s:%d: doOffensiveMove. ball owner is our. press"
                                    ,__FILE__, __LINE__ );
                agent->debugClient().addMessage( "OffPress" );
                Bhv_Press( home_pos ).execute( agent );
                return;
            }
        }
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doOffensiveMove. change to basic move"
                        ,__FILE__, __LINE__ );
    agent->debugClient().addMessage( "OffMove" );
    Bhv_BasicMove( home_pos ).execute( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doDefensiveMove( rcsc::PlayerAgent * agent,
                                    const rcsc::Vector2D & home_pos )
{
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doDefensiveMove"
                        ,__FILE__, __LINE__ );

    //----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    const rcsc::WorldModel & wm = agent->world();

    // intercept
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();
    int opp_min = wm.getInterceptTable()->getOpponentReachCycle();

    if ( self_min < 16 )
    {
        rcsc::Vector2D get_pos
            = rcsc::inertia_n_step_point( wm.ball().pos(),
                                          wm.ball().vel(),
                                          self_min,
                                          rcsc::ServerParam::i().ballDecay() );
        bool enough_stamina = true;
        if ( self_min >= 10
             && wm.self().stamina() < rcsc::ServerParam::i().staminaMax() * 0.7 )
        {
            enough_stamina = false;
        }

        if ( enough_stamina
             && ( home_pos.dist( get_pos ) < 10.0
                  || ( home_pos.absY() < get_pos.absY()
                       && home_pos.y * get_pos.y > 0.0 ) // same side
                  || get_pos.x < home_pos.x
                  )
             && ( opp_min < 3 || wm.existKickableOpponent() )
             )
        {
            agent->debugClient().addMessage( "DefGetBall" );
            rcsc::Body_Intercept().execute( agent );
            agent->setNeckAction( new rcsc::Neck_TurnToBall() );
            return;
        }
    }

    if ( self_min < 15
         && self_min < mate_min + 2
         && ! wm.existKickableTeammate() )
    {
        agent->debugClient().addMessage( "DefGetBall(2)" );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }


    rcsc::Vector2D target_point = home_pos;
    if ( wm.existKickableTeammate()
         || ( mate_min < 2 && opp_min > 2 )
         )
    {
        target_point.x += 10.0;
    }
#if 1
    else if ( home_pos.y * wm.ball().pos().y > 0.0 ) // same side
    {
        target_point.x = wm.ball().pos().x + 1.0;
    }
#endif
    Bhv_BasicMove( target_point ).execute( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
RoleOffensiveHalf::doShootSupportMove( rcsc::PlayerAgent * agent,
                                       const rcsc::Vector2D & home_pos )
{
#if 0
    const rcsc::WorldModel & wm = agent->world();
    if ( home_pos.y * wm.ball().pos().y > 0.0 )
    {
        return false;
    }

    // go to support point
    const double max_x = 39.0;
    const double max_y = 0.0;
    const double att_x = 0.3;
    const double att_y = 0.3;

    rcsc::Vector2D new_home_pos = home_pos;

    new_home_pos.x += wm.ball().pos().x * att_x;
    new_home_pos.y += wm.ball().pos().y * att_y;

    if ( new_home_pos.x > max_x ) new_home_pos.x = max_x;
    if ( home_pos.y > 0.0 && new_home_pos.y < -max_y ) new_home_pos.y = -max_y;
    if ( home_pos.y < 0.0 && new_home_pos.y > max_y ) new_home_pos.y = max_y;

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: shoot support pos(%f, %f)"
                        ,__FILE__, __LINE__,
                        new_home_pos.x, new_home_pos.y );
    agent->debugClient().addMessage( "ShotSupport" );
    Act_AttackerOffensiveMove( new_home_pos ).execute( agent );
#else
    agent->debugClient().addMessage( "ShotSupport" );
    Bhv_AttackerOffensiveMove( home_pos ).execute( agent );
#endif
    return true;
}


/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doShootAreaMove( rcsc::PlayerAgent * agent,
                                    const rcsc::Vector2D & home_pos )
{
    //----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    const rcsc::WorldModel & wm = agent->world();
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: doShootAreaMove"
                        ,__FILE__, __LINE__ );

    if ( wm.getInterceptTable()->isSelfFastestPlayer() )
    {
        if ( rcsc::Body_Intercept().execute( agent ) )
        {
            if ( wm.ball().distFromSelf()
                 < rcsc::ServerParam::i().visibleDistance() )
            {
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            }
            else
            {
                agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            }
            return;
        }
    }

    // support shoot
    if ( doShootSupportMove( agent, home_pos ) )
    {
        return;
    }

    // chase ball
    int self_min = wm.getInterceptTable()->getSelfReachCycle();
    int mate_min = wm.getInterceptTable()->getTeammateReachCycle();

    if ( self_min < mate_min )
    {
        agent->debugClient().addMessage( "GetBall" );
        rcsc::Body_Intercept().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
        return;
    }

    if ( wm.existKickableOpponent()
         && ( wm.getTeammatesFromBall().empty()
              || wm.getTeammatesFromBall().front()->distFromBall() > 2.0 )
         )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: ball owner is opponent. change to normal move"
                            ,__FILE__, __LINE__ );
        Bhv_BasicMove( home_pos ).execute( agent );
        return;
    }

    // set dash power
    static bool S_recover_mode = false;
    double dash_power;
    if ( wm.self().stamina()
         < rcsc::ServerParam::i().recoverDecThrValue() + 300.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > rcsc::ServerParam::i().staminaMax() * 0.7 )
    {
        S_recover_mode = false;
    }

    // recover
    if ( S_recover_mode )
    {
        rcsc::dlog.addText( rcsc::Logger::ROLE,
                            "%s:%d: recovering"
                            ,__FILE__, __LINE__ );
        dash_power
            = wm.self().playerType().staminaIncMax() * wm.self().recovery()
            - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;
    }
    else
    {
        dash_power
            = wm.self().getSafetyDashPower( rcsc::ServerParam::i().maxPower() );
    }

    rcsc::Vector2D target_point = home_pos;
    // check opponent
    if ( ! wm.getOpponentsFromSelf().empty()
         && wm.getOpponentsFromSelf().front()->pos().dist( home_pos ) < 3.0
         && wm.self().pos().dist( home_pos ) < 8.0 )
    {
        if ( home_pos.y > 0.0 ) target_point.y -= 5.0;
        if ( home_pos.y < 0.0 ) target_point.y += 5.0;
    }

    // move
    agent->debugClient().addMessage( "ShootMove%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s:%d: chance area move to (%f, %f) power=%f"
                        ,__FILE__, __LINE__,
                        target_point.x, target_point.y, dash_power );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    if ( ! rcsc::Body_GoToPoint( target_point, dist_thr, dash_power
                                 ).execute( agent ) )
    {
        // face to front or side
        rcsc::Vector2D body_point( 100.0,
                                   wm.self().pos().y );
        if ( wm.ball().angleFromSelf().abs() > 100.0 )
        {
            body_point.x = wm.self().pos().x;
            body_point.y = ( wm.ball().pos().y > wm.self().pos().y
                             ? 100.0
                             : -100.0 );
        }
        rcsc::Body_TurnToPoint( body_point ).execute( agent );
    }

    if ( wm.ball().distFromSelf() < rcsc::ServerParam::i().visibleDistance() )
    {
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
    }
    else
    {
        agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
    }
}

//  LocalWords:  ScanField
